################################################################################
#                                                                              #
#                           DARTFlo viscous module file                        #
#                        Time solver : Time discretization                     #
#                       and computation towards steady state                   #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------

from dart.viscous.Physics.DClosures import Closures
from fwk.coloring import ccolors
import numpy as np

class Solver:
  """Performs the downstream marching process. Calls the pseudo-time integration based
  on Newton's method on demand.
  
  Attributes
  ----------
  
  - integrator (class type): Module able to perform pseudo-time integration of one point.
  - transSolver (class type): Module able to treat the transition related computation.
  - wakeBC (class type): Wake boundary condition.
  - initSln (bool): Flag to initialize the initial condition or to use the value in input.
  """

  def __init__(self, _integrator, transition, wakeBC) -> None:

    # Initialize time parameters
    
    self.integrator = _integrator
    self.transSolver = transition
    self.wakeBC = wakeBC
    self.initSln = 0

  def Run(self, DFlow):
    """Runs the downstream marching solver.
    """

    # Retreive parameters.
    
    nMarkers = DFlow.nN                           # Number of points in the computational domain.
    bNodes = DFlow.bNodes                         # Boundary nodes of the computational domain.
    convergedPoints = -1*np.ones(DFlow.nN)        # Convergence control of each point.
    convergedPoints[0] = 0                        # Boundary condition.
    
    self.transSolver.initTransition(DFlow)        # Initialize the flow regime on each cell.

    lockTrans = 0                                 # Flag to remember if the transition is already found or not.
    
    for iMarker in range(1, nMarkers):
      displayState = 0                            # Flag to output simulation state.

      if self.initSln:
        iMarkerPrev = 0 if iMarker == DFlow.bNodes[1] else iMarker - 1
        DFlow.u[iMarker*DFlow.nVar: (iMarker + 1)*DFlow.nVar] = DFlow.u[iMarkerPrev*DFlow.nVar: (iMarkerPrev + 1)* DFlow.nVar]
      if DFlow.flowRegime[iMarker] == 1 and DFlow.u[iMarker * DFlow.nVar + 4] == 0:
        DFlow.u[iMarker * DFlow.nVar + 4] = 0.03

      if iMarker == bNodes[0] + 1:	# Upper side start.
         print('- Upper side,', end = ' ')
      
      if iMarker == bNodes[1]:	# Lower side start.
        if lockTrans == 0:
          print('no transition.')
        lockTrans = 0
        print('- Lower side,', end = ' ')

      if iMarker == bNodes[2]: # First wake point

        self.wakeBC(DFlow)	# Wake boundary condition.
        convergedPoints[iMarker] = 0
        if lockTrans == 0:
          print('no transition.')
        lockTrans = 1				

        continue	# Ignore remaining instructions for the first wake point.
      
      # Call Newton solver for the point.

      convergedPoints[iMarker] = self.integrator.TimeMarching(DFlow, iMarker, displayState)

      if convergedPoints[iMarker] != 0:
        print(ccolors.ANSI_RED + 'Marker {:<4.0f} (x/c={:<1.4f}): PTI terminated with exit code {:<4.0f}.'
        .format(iMarker, DFlow.xGlobal[iMarker], convergedPoints[iMarker]), ccolors.ANSI_RESET)

      # Check for transition.

      if not lockTrans:

        self.__CheckWaves(DFlow, iMarker)
        
        # Free transition.
        if DFlow.u[iMarker * DFlow.nVar + 2] >= DFlow.Ncrit:
          self.__AverageTransition(DFlow, iMarker, displayState)
          if iMarker < DFlow.bNodes[1]:
            print('free transition x/c = {:<1.5f} {:>4.0f}.'.format(DFlow.xtr[0],iMarker))
          elif iMarker < DFlow.bNodes[2]:
            print('free transition x/c = {:<1.5f} {:>4.0f}.'.format(DFlow.xtr[1],iMarker))
          else:
            print('error in wake')
          lockTrans = 1
        
        # Forced transition.
        elif DFlow.xtrF[0] is not None or DFlow.xtrF[1] is not None:
          if iMarker != 1 and DFlow.flowRegime[iMarker] == 1 and DFlow.flowRegime[iMarker - 1] == 0:
            print('forced transition near x/c = {:<2.3f} {:>4.0f}'.format(DFlow.xGlobal[iMarker],iMarker))
            self.__AverageTransition(DFlow, iMarker, displayState, forced = 1)
            lockTrans = 1

    return convergedPoints

  def __CheckWaves(self, var, iMarker) -> None:
    """Determine if the amplification waves start growing or not.
    """

    logRet_crit = 2.492*(1/(var.blPar[(iMarker)*var.nBlPar + 6]-1))**0.43
    + 0.7*(np.tanh(14*(1/(var.blPar[(iMarker)*var.nBlPar + 6]-1))-9.24)+1.0)

    if var.blPar[(iMarker)*var.nBlPar + 0] > 0: # Avoid log of negative number.
      if np.log10(var.blPar[(iMarker)*var.nBlPar + 0]) <= logRet_crit - 0.08:
        var.u[iMarker*var.nVar + 2] = 0
    
    pass


  def __AverageTransition(self, var, iMarker, displayState, forced = 0) -> None:
    """Averages laminar and turbulent solution at the transition location.
    The boundary condition of the turbulent flow portion is also imposed.
    """

    # Save laminar solution.
    lamSol = var.u[(iMarker)*var.nVar : (iMarker+1)* var.nVar].copy()

    # Set up turbulent portion boundary condition.
    if not forced:
      avTurb = self.transSolver.solveTransition(var, iMarker)
    else:
      avTurb = self.transSolver.forcedTransition(var, iMarker)

    # Compute turbulent solution.
    var.flowRegime[iMarker] = 1
    iMarkerPrev = 0 if iMarker == var.bNodes[1] else iMarker - 1

    # Avoid starting with 0 for the shear stress and high H.
    var.u[iMarker*var.nVar + 4] = var.u[iMarkerPrev*var.nVar + 4]
    var.u[iMarker*var.nVar + 1] = 1.515

    self.integrator.TimeMarching(var, iMarker, displayState)

    # Average solutions.
    avLam      = 1 - avTurb
    thetaTrans = avLam * lamSol[0] + avTurb * var.u[(iMarker)*var.nVar + 0]
    HTrans     = avLam * lamSol[1] + avTurb * var.u[(iMarker)*var.nVar + 1]
    nTrans     = 	 0 
    ueTrans    = avLam * lamSol[3] + avTurb * var.u[(iMarker)*var.nVar + 3]
    CtTrans    = avTurb * var.u[(iMarker)*var.nVar + 4]

    var.u[(iMarker)*var.nVar : (iMarker+1)*var.nVar] = [thetaTrans, HTrans, nTrans, ueTrans, CtTrans]

    # Recompute closures.
    var.blPar[(iMarker)*var.nBlPar+1:(iMarker)*var.nBlPar+9]=Closures.turbulentClosure(var.u[(iMarker)*var.nVar+0],
                                            var.u[(iMarker)*var.nVar+1],
                                            var.u[(iMarker)*var.nVar+4],
                                            var.blPar[(iMarker)*var.nBlPar+0],
                                            var.extPar[(iMarker)*var.nExtPar+0],
                                            'airfoil')
    var.blPar[(iMarker-1)*var.nBlPar+1:(iMarker-1)*var.nBlPar+7]=Closures.laminarClosure(var.u[(iMarker-1)*var.nVar+0],
                                            var.u[(iMarker-1)*var.nVar+1],
                                            var.blPar[(iMarker-1)*var.nBlPar+0],
                                            var.extPar[(iMarker-1)*var.nExtPar+0],
                                            'airfoil')
    pass